#version 130
#extension GL_ARB_texture_rectangle : enable
#define REZ 8
#define REZ2 (REZ * REZ)
#define FLT_MAX 3.4028235e+038
#define STACK_MAX 25

uniform sampler2DRect u_texture_position;
uniform sampler2DRect u_texture_hierarchy;
uniform int u_hierarchy_texWidth;
uniform int u_num_leaf_nodes;
uniform float u_dist_max;

vec4 getHierarchyPosition(int index) {
    float i = mod(index, u_hierarchy_texWidth);
    float j = float (index / u_hierarchy_texWidth);
    return texture2DRect(u_texture_hierarchy, vec2(4 * i + 0.5, j + 0.5));
}
vec4 getHierarchyNormal(int index) {
    float i = mod(index, u_hierarchy_texWidth);
    float j = float (index / u_hierarchy_texWidth);
    return texture2DRect(u_texture_hierarchy, vec2(4 * i + 1.5, j + 0.5));
}
vec4 getHierarchyColorf(int index) {
    float i = mod(index, u_hierarchy_texWidth);
    float j = float (index / u_hierarchy_texWidth);
    return texture2DRect(u_texture_hierarchy, vec2(4 * i + 2.5, j + 0.5));
}
vec4 getHierarchyColorb(int index) {
    float i = mod(index, u_hierarchy_texWidth);
    float j = float (index / u_hierarchy_texWidth);
    return texture2DRect(u_texture_hierarchy, vec2(4 * i + 3.5, j + 0.5));
}

vec3 calcProjection(vec3 face[3], vec3 v) {
    vec3 d0 = face[0] - face[2];
    vec3 d1 = face[1] - face[2];
    vec3 e = v - face[2];
    float a00 = dot(d0, d0);
    float a11 = dot(d1, d1);
    float a01 = dot(d0, d1);
    vec2 b = vec2(dot(d0, e), dot(d1, e));
    mat2 invA = mat2(a11, -a01, -a01, a00);
    invA /= a00 * a11 - a01 * a01;
    vec2 lambda = invA * b;
    return lambda[0] * face[0] + lambda[1] * face[1] + (1.0 - lambda[0] - lambda[1]) * face[2];
}

extern float areaWeight[REZ2];

void main() {
    vec3 position = texture2DRect(u_texture_position , gl_FragCoord.xy).xyz;
    
    // initialize cube map micro-buffer
    uint buffer[REZ2 * 6];
    
    // cube map orientation
    vec3 viewDirection[6] = vec3[6] (vec3(1, 0, 0), vec3(0, -1, 0), vec3(-1, 0, 0), vec3(0, 1, 0), vec3(0, 0, 1), vec3(0, 0, -1));
    vec3 upDirection  [6] = vec3[6] (vec3(0, 0, 1), vec3(0,  0, 1), vec3( 0, 0, 1), vec3(0, 0, 1), vec3(0, 1, 0), vec3(0, 1,  0));
    
    const uint initial_value = 0x3FFu;
    for (int i = 0; i < REZ2 * 6; ++i)
        buffer[i] = initial_value;
    
    // rendering by traversing hierarchy
    int index_stack[STACK_MAX];// = int[STACK_MAX] (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    index_stack[0] = 0;
    int stack_len = 1;
    int stack_len_max = 0;
    while (0 < stack_len) {
        stack_len_max = max(stack_len_max, stack_len);
        int index = index_stack[stack_len - 1];
        --stack_len;
        vec3 node_position = getHierarchyPosition(index).xyz;
        vec3 theta_phi_rad = getHierarchyNormal  (index).xyz;
        vec3 node_colorf   = getHierarchyColorf  (index).xyz;
        vec3 node_colorb   = getHierarchyColorb  (index).xyz;
        float theta = theta_phi_rad.x;
        float phi   = theta_phi_rad.y;
        vec3  node_normal  = vec3(sin(phi) * cos(theta), sin(phi) * sin(theta), cos(phi));
        float node_radius = theta_phi_rad.z;
        vec3 v = node_position - position;
        int buffer_id = -1;
        float dot_max = -FLT_MAX;
        for (int i = 0; i < 6; ++i) {
            float view_v = dot(viewDirection[i], v);
            if (dot_max < view_v) {
                dot_max = view_v;
                buffer_id = i;
            }
        }
        vec3 e0 = viewDirection[buffer_id];
        vec3 e2 = upDirection  [buffer_id];
        vec3 e1 = cross(e0, e2);
        vec3 d = v / dot(v, e0);
        float t1 = dot(d, e1);
        float t2 = dot(d, e2);     // -1 <= t1, t2 <= 1
        int ix = min(int((t1 + 1.0) * float(REZ) * 0.5), REZ - 1);
        int iy = min(int((t2 + 1.0) * float(REZ) * 0.5), REZ - 1);     // 0 <= ix, iy <= rez - 1
        float s1 = (float(ix) - float(REZ) * 0.5) / (float(REZ) * 0.5);
        float s2 = (float(iy) - float(REZ) * 0.5) / (float(REZ) * 0.5);               // -1 <= s1, s2 <= 1 - 2 / rez_
        vec3 h[4];
        h[0] = e0 +  s1                     * e1 +  s2                     * e2;
        h[1] = e0 + (s1 + 2.0 / float(REZ)) * e1 +  s2                     * e2;
        h[2] = e0 + (s1 + 2.0 / float(REZ)) * e1 + (s2 + 2.0 / float(REZ)) * e2;
        h[3] = e0 +  s1                     * e1 + (s2 + 2.0 / float(REZ)) * e2;
        vec4 pixelEdgeDist = vec4(0);
        float max_stickout = 0;
        for (int i = 0; i < 4; ++i) {
            vec3 face[3] = vec3[3] (vec3(0), h[i], h[i == 3 ? 0 : (i + 1)]);
            vec3 projection = calcProjection(face, v);
            pixelEdgeDist[i] = length(projection - v);
            max_stickout = max(0, node_radius - pixelEdgeDist[i]);
        }
        if (0 < max_stickout && index < u_num_leaf_nodes - 1) {       // if current node sticks out of this pixel and has children
            ++stack_len;
            index_stack[stack_len - 1] = 2 * index + 2;
            ++stack_len;
            index_stack[stack_len - 1] = 2 * index + 1;
        } else {
            float pixel_size = 0.5 * (pixelEdgeDist[0] + pixelEdgeDist[1] + pixelEdgeDist[2] + pixelEdgeDist[3]);
            int k = int(ceil(max_stickout / pixel_size));
            uint node_dist = uint(float(0x3FF) * length(v) / u_dist_max);
            bool node_isFront = dot(node_normal, v) < 0;
            int buffer_offset = REZ2 * buffer_id;
            const int k_max = 3;
            for (int dy = -k_max; dy <= k_max; ++dy) {
                if (k < abs(dy))
                    continue;
                for (int dx = -k_max; dx <= k_max; ++dx) {
                    if (k < abs(dx))
                        continue;
                    int ix1 = ix + dx;
                    int iy1 = iy + dy;
                    if (ix1 < 0 || REZ <= ix1 || iy1 < 0 || REZ <= iy1)
                        continue;
                    int buffer_index = buffer_offset + iy1 * REZ + ix1;
                    if (node_dist < (buffer[buffer_index] & 0x3FFu))
                        buffer[buffer_index] = (uint(index) << 11u) | ((node_isFront ? 1u : 0u) << 10u) | node_dist;
                }
            }
        }
    }
    // integration
    vec3 result = vec3(0);
    float weight_sum = 0;
    for (int i = 0; i < REZ2 * 6; ++i) {
        float node_dist = u_dist_max * float(buffer[i] & 0x3FFu) / float(0x3FF);
        int j = int(mod(i, REZ2));
        float weight = areaWeight[j] / node_dist;
        bool node_isFront = (buffer[i] & 0x400u) != 0u;
        int node_index = int(buffer[i] >> 11u);
        vec3 node_color = node_isFront ? getHierarchyColorf(node_index).xyz : getHierarchyColorb(node_index).xyz;
        result += weight * node_color;
        weight_sum += weight;
    }
    result /= weight_sum;

    gl_FragColor = vec4(result, 1);
}
